package IBL;
use strict;
use warnings;

#////////////////////////////////////////////////////////////////
#/								/
#/   IBL.pm							/
#/								/
#/   A collection of Perl routines to ease development of       /
#/   application using iButtonLink 1-wire masters.		/
#/								/
#/   Copyright 2010 ... All rights reserved.                    /
#/   								/
#/   iButtonLink LLC						/
#/   N8921 Stone School Rd.					/
#/   East Troy, Wi. 53120					/
#/   Ph: (262)662-4029						/
#/								/
#/								/
#/   Relase date: 30-March-2010                                 /
#/  
#/   Fixed some bugs in Read26_TADC() function by 
#/   Moritz Fuchs Elektronik - May, 2012
#/                                                              /
#////////////////////////////////////////////////////////////////

BEGIN {
    # This must be in a BEGIN block for 'use' to be conditional
    if ($^O eq 'MSWin32') {
        eval "use Win32::SerialPort";
    } else {
        eval "use Device::SerialPort";
    }
    die "No SerialPort module available: $@\n" if ($@);
}


use Net::Telnet ();
use Time::HiRes qw(usleep gettimeofday);

#//////////////////////////////////////
#/   Table used by crc8 routines to   /
#/   calculate CRC8 validations.      /
#/   The table is populated here so   /
#/   it is only done once.            /
#//////////////////////////////////////
my @crc8_table = (
   0x00, 0x5e, 0xbc, 0xe2, 0x61, 0x3f, 0xdd, 0x83, 
   0xc2, 0x9c, 0x7e, 0x20, 0xa3, 0xfd, 0x1f, 0x41, 
   0x9d, 0xc3, 0x21, 0x7f, 0xfc, 0xa2, 0x40, 0x1e, 
   0x5f, 0x01, 0xe3, 0xbd, 0x3e, 0x60, 0x82, 0xdc, 
   0x23, 0x7d, 0x9f, 0xc1, 0x42, 0x1c, 0xfe, 0xa0, 
   0xe1, 0xbf, 0x5d, 0x03, 0x80, 0xde, 0x3c, 0x62, 
   0xbe, 0xe0, 0x02, 0x5c, 0xdf, 0x81, 0x63, 0x3d, 
   0x7c, 0x22, 0xc0, 0x9e, 0x1d, 0x43, 0xa1, 0xff, 
   0x46, 0x18, 0xfa, 0xa4, 0x27, 0x79, 0x9b, 0xc5, 
   0x84, 0xda, 0x38, 0x66, 0xe5, 0xbb, 0x59, 0x07, 
   0xdb, 0x85, 0x67, 0x39, 0xba, 0xe4, 0x06, 0x58, 
   0x19, 0x47, 0xa5, 0xfb, 0x78, 0x26, 0xc4, 0x9a, 
   0x65, 0x3b, 0xd9, 0x87, 0x04, 0x5a, 0xb8, 0xe6, 
   0xa7, 0xf9, 0x1b, 0x45, 0xc6, 0x98, 0x7a, 0x24, 
   0xf8, 0xa6, 0x44, 0x1a, 0x99, 0xc7, 0x25, 0x7b, 
   0x3a, 0x64, 0x86, 0xd8, 0x5b, 0x05, 0xe7, 0xb9, 
   0x8c, 0xd2, 0x30, 0x6e, 0xed, 0xb3, 0x51, 0x0f, 
   0x4e, 0x10, 0xf2, 0xac, 0x2f, 0x71, 0x93, 0xcd, 
   0x11, 0x4f, 0xad, 0xf3, 0x70, 0x2e, 0xcc, 0x92, 
   0xd3, 0x8d, 0x6f, 0x31, 0xb2, 0xec, 0x0e, 0x50, 
   0xaf, 0xf1, 0x13, 0x4d, 0xce, 0x90, 0x72, 0x2c, 
   0x6d, 0x33, 0xd1, 0x8f, 0x0c, 0x52, 0xb0, 0xee, 
   0x32, 0x6c, 0x8e, 0xd0, 0x53, 0x0d, 0xef, 0xb1, 
   0xf0, 0xae, 0x4c, 0x12, 0x91, 0xcf, 0x2d, 0x73, 
   0xca, 0x94, 0x76, 0x28, 0xab, 0xf5, 0x17, 0x49, 
   0x08, 0x56, 0xb4, 0xea, 0x69, 0x37, 0xd5, 0x8b, 
   0x57, 0x09, 0xeb, 0xb5, 0x36, 0x68, 0x8a, 0xd4, 
   0x95, 0xcb, 0x29, 0x77, 0xf4, 0xaa, 0x48, 0x16, 
   0xe9, 0xb7, 0x55, 0x0b, 0x88, 0xd6, 0x34, 0x6a, 
   0x2b, 0x75, 0x97, 0xc9, 0x4a, 0x14, 0xf6, 0xa8, 
   0x74, 0x2a, 0xc8, 0x96, 0x15, 0x4b, 0xa9, 0xf7, 
   0xb6, 0xe8, 0x0a, 0x54, 0xd7, 0x89, 0x6b, 0x35); 

sub new {
    my $classname = shift;
    my $self = bless {}, $classname;
    $self->{DEVICE}="67.53.90.115:10001";        #Default IBL master communications path
    $self->{ANSWER}="";
    $self->{IBLink}="";                          # Populated by new
    $self->{OS}=$^O;                             # Default Operating System
    $self->{INTERFACE}="Telnet";                 # Default interface type Telnet
    $self->{TEMP_SCALE}="C";                     # Default is American (Farenheit)
    $self->{IBLMASTER} = undef;                  # Default is "Don't know"
    $self->{OWRESET_RESULT} = undef;             # Populated by OW_Reset

    return $self;
}
#/////////////////////////
#/   Temperature scale   /
#/   F = Farenheit       /
#/   C = Celcius         /
#/   K = Kelvin          /
#/////////////////////////
sub temp_scale {
    my $self = shift;
    if(@_) { $self->{TEMP_SCALE} = shift }
    return $self->{TEMP_SCALE};
}
#/////////////////////////////////////////
#/   Communications Channel ... DEVICE   /
#/   If INTERFACE is Telnet, this should /
#/   be the IP address:Port of the master/
#/   (LinkHub-E)                         /
#/                                       /
#/   If interface is Serial, this should /
#/   be the Serial Device name           /
#/      If OS = Win then this should be  /
#/           in the form \\.\COMx        /
#/      If OS = Linux this should be in  /
#/           the form /dev/ttyUSBx       /
#/                                       /
#/////////////////////////////////////////
sub device {
    my $self = shift;
    if(@_) { $self->{DEVICE} = shift }
    return $self->{DEVICE};
}
#////////////////////////
#/   Operating System   /
#/   Windows = Win      /
#/   Linux = Linux      /
#/   VMS = VMS          /
#////////////////////////
sub os {
    my $self = shift;
    if(@_) { $self->{OS} = shift }
    return $self->{OS};
}

sub interface {
    my $self = shift;
    if(@_) { $self->{INTERFACE} = shift }
    return $self->{INTERFACE};
}

sub master {
    my $self = shift;
    if(@_) { $self->{IBLMASTER} = shift }
    return $self->{IBLMASTER};
}
sub reset_result {
    my $self = shift;
    return $self->{OWRESET_RESULT}                # Only set by OW_Reset
}
#//////////////////////////////////////////
#/                                        /
#/   Create a Classified Inventory hash   /
#/   of all the sensors on the bus        /
#/                                        /
#/   The key of the returned hash is the  /
#/   serial number of the sensor.         /
#/   The value of the hash is a comma     /
#/   delimited string,  the first entry   /
#/   of which is the general sensor type  /
#/   example "T" ... Temperature          /
#/      "TH" ... Temperature, Humidity    /
#/                                        /
#/   The hash entry is ready to be used   /
#/   by Read_Sensor.                      /
#/                                        /
#/   If the hash entry is a composite     /
#/   sensor (eg T-String) the remainder   /
#/   of the hash value is information     /
#/   needed to process the composite      /
#/                                        /
#//////////////////////////////////////////
sub ClassifyedInventory {
    my $self = shift;
    my @SN = $self->RawInventory();       #Get the serial number of all the 1-wire devices
    my %hinv;                             #A local empty hash to return
    my ($x, $y, $key, $value) = 0;
    my @x;
    
    
    foreach (@SN) {

        my $SerialNumber = $_;

        if (substr($SerialNumber,14,2) eq "2D") {
            $x = $self->Expand_TString($SerialNumber);
            $hinv{$SerialNumber} = "TS," . $x;
            next;
        }
        if (substr($SerialNumber,14,2) eq "01") { $hinv{$SerialNumber} = "SN"; next }
        if (substr($SerialNumber,14,2) eq "28") { $hinv{$SerialNumber} = "T"; next }
        if (substr($SerialNumber,14,2) eq "10") { $hinv{$SerialNumber} = "T"; next }
        if (substr($SerialNumber,14,2) eq "26") {
            $x = $self->Read26_page($SerialNumber,"03");
            if(substr($x,0,2) eq "00") { $hinv{$SerialNumber} = "T"; }
            if(substr($x,0,2) eq "19") {
                if(substr($x,2,2) eq "00") { $hinv{$SerialNumber} = "TH,3610"; }
                if(substr($x,2,2) eq "01") { $hinv{$SerialNumber} = "TH,4000"; }
            }
            if(substr($x,0,2) eq "1A") { $hinv{$SerialNumber}="TV"; }
            if(substr($x,0,2) eq "1B") { $hinv{$SerialNumber}="TL"; }
            if(substr($x,0,2) eq "1C") { $hinv{$SerialNumber}="TC"; }
            if(substr($x,0,2) eq "1D") { $hinv{$SerialNumber}="TW"; }
            if(substr($x,0,2) eq "1E") { $hinv{$SerialNumber}="THW"; }
            if(substr($x,0,2) eq "1F") { $hinv{$SerialNumber}="TPS"; }
            if(substr($x,0,2) eq "20") { $hinv{$SerialNumber}="TPD"; }
            next;
        }
        $hinv{$SerialNumber} = "UN";
    }
   # Now remove the DS18B20's that belong to a T-String
    while (($key, $value) = each(%hinv)){
        if (substr($value,0,2) eq "TS") {
            @x = split(',',$value);        # form an array from comman delimited string
            shift(@x);                     # get rid of first element
            foreach $y (@x) {
                delete($hinv{$y});          # Delete the hash for the string 18b20
            }
        }
    }
    return %hinv;
}
#/////////////////////////////////////////////////////////////////
#/                                                               /
#/   Open the communications channel to the iButtonLink master   /
#/                                                               /
#/////////////////////////////////////////////////////////////////
sub open {
    my $self = shift;
    my $ok = "";                            # Status returns
    my $retry = 0;                          # Retry variable

    if (($self->{INTERFACE}) eq "Serial") {      #if the interface type is Serial
        if ($^O eq "MSWin32") {
            $self->{IBLink} = new Win32::SerialPort($self->{DEVICE}) || die "Win32 Could not open Link master $self->{DEVICE}\n";
        } else {
            $self->{IBLink} = new Device::SerialPort($self->{DEVICE}) || die "linux Could not open Link master $self->{DEVICE}\n";
	}
            $self->{IBLink}->baudrate(9600);
            $self->{IBLink}->stopbits(1);
            $self->{IBLink}->databits(8);
            $self->{IBLink}->parity('none');
            $self->{IBLink}->handshake('none');
            $self->{IBLink}->write_settings;        
            usleep (200_000);                   #Wait 200 milliseconds for classic Link to power up
            $self->{IBLink}->purge_all();       # Empty input buffer
    }
    if ($self->{INTERFACE} eq "Telnet") {
        my @y = split(":",$self->{DEVICE});
        if (!exists $y[1]) { $y[1] = 10001; } #Default to 10001 if port not specified
        $self->{IBLink} = new Net::Telnet (Binmode => 1,
                                           Timeout => 2);
        $ok = $self->{IBLink}->open(Host => $y[0], Port => $y[1]);
        if (!$ok) {
            print "Failed to create telnet session to $y[0] port $y[1] return $ok\n";
            exit(0);
        }
        $self->{IBLink}->buffer_empty(); # Belt and suspenders ... should be nothing in input buffer
    }
    while ($retry < 3){
        $ok = $self->setup_master();     # Probe for a iButtonLink master on the communications path
        if ($ok) {return}                # Found one ... continue
        $ok = $self->reset_master();     # Did not find one ... try to reset it to command mode
        $retry++;
    }
}
#////////////////////////////////////////////////////
#/                                                  /
#/   Write a string to the communications channel   /
#/                                                  /
#////////////////////////////////////////////////////
sub write {
    my $self = shift;
    my $x = shift;                       #String passed in to write
    my $ok = 0;                          # Status return
    my $xlen = length($x);
    if($self->{INTERFACE} eq "Serial") {
        $ok = $self->{IBLink}->write($x);
        if ($ok != $xlen) {
            print "Write failed to Serial object\n";
        }
    }
    if($self->{INTERFACE} eq "Telnet") {
        $ok = $self->{IBLink}->put($x);
        if ($ok != 1) { print "Write failed to Telnet object\n" }
    }
}

#///////////////////////////////////////////////////////////////////
#/                                                                 /
#/   reset_master ... try to get an IBL master into command mode   /
#/                                                                 /
#/   possible reasons for calling this routine.                    /
#/      1. master was left in binary mode by a previous connection /
#/      2. master is not communicating at the same speed as we are /
#/      3. master did not respond during setup_master.             /
#/                                                                 /
#///////////////////////////////////////////////////////////////////
sub reset_master {
    my $self=shift;
    return;
}
#///////////////////////////////////////////////////////
#/                                                     /
#/   Read number of bytes from communications channel  /
#/                                                     /
#/  Example: $obj->read(10);  # Read 10 bytes          /
#/                                                     /
#///////////////////////////////////////////////////////
sub read {
#///////////////////////////////////
#/   Serial timeouts not working   /
#///////////////////////////////////
    my $self = shift;                   # Requested number of characters to read
    my $bytes_wanted = 1;               # Default to 1
    if (@_) { $bytes_wanted = shift; }  # retrieve number of bytes to read if specified
    my $z = "";                         # a local answer string
    my $x = "";                         # data from lower level protocol
    my $count_in = 0;                   # count of bytes returned from lower level
    my $bytes_in = 0;                   # The data returned from lower level

    while(1) {
        if($self->{INTERFACE} eq "Serial") {
            ($count_in, $bytes_in) = $self->{IBLink}->read($bytes_wanted);
            $bytes_wanted -= $count_in;
            $z .= $bytes_in;        
            if ($bytes_wanted <= 0) { return $z }
        }

        if($self->{INTERFACE} eq "Telnet") {
            while(1) {
                $bytes_in = $self->{IBLink}->get();
                $bytes_wanted -= length($bytes_in);
                $z .= $bytes_in;        
                if ($bytes_wanted <= 0) { return $z }
            }
        }
    }
}

sub readline {
#///////////////////////////////////
#/   Serial timeouts not working   /
#///////////////////////////////////
    my $self = shift;
    my ($x,$z,$y) ="";

    if ($self->{INTERFACE} eq "Telnet") {
        $x = $self->{IBLink}->getline(Timeout=>2);
        if (!defined($x)) {
            print "getline returned nothing\n";
        }
        return $x;
    }

    if ($self->{INTERFACE} eq "Serial") {
#        $self->{ANSWER}=$self->{IBLink}->streamline("\n");
#        return;
#    }
        while(1) {
            ($x,$y) = $self->read(1);                #Read next character from lower protocol
            if ($x eq chr(0x0A)) {
                $z .= $x;
                return $z;
            } else {
                $z .= $x;
            }
        }
    }
    $self->{ANSWER} = undef;
    return $self->{ANSWER};
}
#////////////////////////////////////////////////////////
#/                                                      /
#/   setup_master ... Identify the iButtonLink master   /
#/                    and optimize parameters for       /
#/                    that particular master.           /
#/                                                      /
#////////////////////////////////////////////////////////
sub setup_master {
    my $self = shift;
    my $x = "";
    $self->write(' ');                      # Ask for master's verision banner
    $x = $self->readline;                   # Read the response.
    chomp($x);                              # Get rid of the line feed
    #///////////////////////////////////////////////////////////////////////////
    #/ This should timeout if pointed to an empty port ... fix it so it does   /
    #///////////////////////////////////////////////////////////////////////////
    if (!defined($x)) { return 1; }           # Didn't get a response
    
    if (index($x,"LinkUSB") != -1) {
        $self->{IBLink}->handshake('rts');  # LinkUSB's can do hardware flow
        $self->{IBLink}->write_settings;
        $self->{IBLMASTER} = $x;            # Save for others to find
        return 1;
    }
    if (index($x,"LinkHub") != -1) {
        $self->{IBLMASTER} = $x;            # Save for others to find
        return 1;
    }

    if (index($x,"LinkOEM") != -1) {
        $self->{IBLMASTER} = $x;            # Save for others to find
        return 1;
    }

    if (index($x,"Link ") != -1) {
        $self->{IBLMASTER} = $x;             # Save for others to find
        return 1;
    }

    $self->{IBLMASTER} = "Unknown - $x";
    return                                  # Did not find a Link Master we recognized
}
#////////////////////////////////////////////////////////////////////////////
#/                                                                          /
#/   RawInventory ... return an array containing all 1-wire serial numbers  /
#/                    present on the bus.                                   /
#/                                                                          /
#////////////////////////////////////////////////////////////////////////////
sub RawInventory {
    my $self = shift;                   # get us going
    my @SN=();                          # An empty array
    my $line = "";
    my $ok;
    my $x = "";                         # Scratchpad
    my $FirstChar = "";
    #////////////////////////////////////////////////////
    #/   Do a reset and check for at least one sensor   /
    #////////////////////////////////////////////////////
    $ok = $self->OW_Reset();       	# Get everyone started clean
    if (!$ok) {
        $x = $self->reset_result();     # get the last OW_Reset result
        if ('N' eq $x) {                # Test for 'N'o presence
            return @SN;                 # Return an empty array
        }
    die;
    }

    $self->write('f');                  # send a first command
    $line = $self->readline();          # Get response
    if ($line eq "N") { return @SN; }   # There are no sensors
    
   #////////////////////////////////////////////////////
   #/   Looping adding serial numbers to array until   /
   #/   the master returns the last serial number      /
   #/   indicator.                                     /
   #////////////////////////////////////////////////////
   while (1) {                       		#While the line contains a +
      $ok = $self->crc8_RL(substr($line,2,16));  #See if the SN CRC checks out
      if (!$ok) {
        $FirstChar = substr($line,0,1);      #Remember if line starts with "+" or "-"
        push (@SN,substr($line,2,16));          #Put serial number into result array
      }
        last if ($FirstChar eq '-');
        $self->write('n');
        $line = $self->readline();
   }
   return @SN;
}
#//////////////////////////////////////////////////
#/                                                /
#/   ReverseSN ... invert the bytes in a 1-wire   /
#/      serial number.  Since the byte of a       /
#/      serial number are represented as a pair   /
#/      of ASCII hex characters, the reverseing   /
#/      is by pairs of ASCII characters.          /
#/                                                /
#//////////////////////////////////////////////////
sub ReverseSN {
    my $self = shift;
    my $xxline = shift;
    my $newsn = substr($xxline,-2,2);
    $newsn = $newsn . substr($xxline,-4,2);
    $newsn = $newsn . substr($xxline,-6,2);
    $newsn = $newsn . substr($xxline,-8,2);
    $newsn = $newsn . substr($xxline,-10,2);
    $newsn = $newsn . substr($xxline,-12,2);
    $newsn = $newsn . substr($xxline,-14,2);
    $newsn = $newsn . substr($xxline,-16,2);
    return $newsn;
}
#////////////////////////////////////////////
#/                                          /
#/   Preform a 1-Wire reset.                /
#/                                          /
#/   The result of the OW_Reset are posted  /
#/   to OWRESET_RESULT so that the calling  /
#/   program can query the result.          /
#/                                          /
#/   Returns 1 (true) if result was "P"     /
#/   0 (false) if any thing else            /
#/                                          /
#/   The result is stored in OWRESET_RESULT /
#/   for retrieval and processing later by  /
#/   the caller.                            /
#/                                          /
#////////////////////////////////////////////
sub OW_Reset {
    my $self = shift;
    my $line = "";
    my $save_line = "";
    my $crlf = chr(0x0D) . chr(0x0A);
    #/////////////////////////////////////////////////////
    #/   Empty the input buffer to resynch data stream   /
    #/////////////////////////////////////////////////////
    if ($self->{INTERFACE} eq "Telnet") { $self->{IBLink}->buffer_empty(); }
    if ($self->{INTERFACE} eq "Serial"){ $self->{IBLink}->purge_all(); }
    
    $self->write("r");                        # Send a "r" to the 1-wire master
    $save_line = $self->readline();           # retrieve response
    
    if (length($save_line) == 3 ) {
        $line = substr($save_line,0,1);
    }
    
    if ($line eq "P") {                       # Process "P"resence response
        $self->{OWRESET_RESULT} = 'P';    
        return 1;
    }
    if ($line eq "N") {                       # Process "N"o presence response
        $self->{OWRESET_RESULT} = 'N';
        return 0;
    }
    
    if ($line eq "S") {                       # Process "S"hort response
        $self->{OWRESET_RESULT} = 'S';
        return 0;
    }

    if ($line eq "E") {                       # Process "E"rror response
        $self->{OWRESET_RESULT} = 'E';
        return 0;
    }
    printf "OW_Reset failed: %v.2X\n",$save_line;  #Display unknown response
    $self->{OWRESET_RESULT} = undef;
    return 0;                                 # Return "false"
}
#////////////////////////////////////////
#/                                      /
#/   Stub routine for exit processing   /
#/                                      /
#////////////////////////////////////////
sub DESTROY {
    my $self = shift;
}
#//////////////////////////////////////////////////////////////
#/                                                            /
#/   Read26_page ... Return a DS2438 page data as ASCII hex   /
#/                                                            /
#/   Input: SerialNumber of DS2438 (ASCII Hex LSB = family)   /
#/          page number to read (ASCII Hex)                   /
#/                                                            /
#//////////////////////////////////////////////////////////////
sub Read26_page {
    my $self = shift;
    my $SerialNumber = shift;           # Get Serial Number passed in
    my $ok = 0;
    my $pageno = shift;                 # Get the page number to read
    my $select = "b55" . $self->ReverseSN($SerialNumber);
    
    $ok = $self->OW_Reset();

    $self->write($select . "B8"); #Recall to scratchpad
    $self->write("$pageno\r");              # the page number to recall

    my $line = $self->readline(); 	# Consume command echo

    $ok = $self->OW_Reset();
    $self->write($select . "BE");
    $self->write($pageno . "FFFFFFFFFFFFFFFFFF\r"); #Read scratchpad
    $line = $self->readline();           #retrieve response w/2 second timeout
    my $data = substr($line,22,18);

    $ok = $self->crc8_LR(substr($data,0,18));   # Validate CRC on data read
    if ($ok != 0) {
       print "CRC on data read from DS2438" . substr($select,3,16) . "\n";
       return undef;
    }
    return $data;
}
#//////////////////////////////////////////////////////////////
#/                                                            /
#/   Write26_page ... Write a page of DS2438 memory           /
#/                    through scratchpad.                     /
#/					                      /
#/   Input: SerialNumber of DS2438 (ASCII Hex LSB = family)   /
#/          8 data bytes to be written (ASCII Hex)            /
#/          page number to write (ASCII Hex)                  /
#/	              					      /
#/   Output: A numberic scalar indicating success (0)         /
#/           or failure (<> 0>.	                              /
#/						              /
#//////////////////////////////////////////////////////////////
sub Write26_page {
    my $self = shift;
    my $SerialNumber = shift;
    my ($y, $data, $ok) = 0;
    my $pagedata = shift;
    if (length($pagedata) > 16) { $pagedata = substr($pagedata,0,16); }
    my $pageno = shift;
    my $select = "b55" . $self->ReverseSN($SerialNumber);

    $self->OW_Reset();                   # Reset the bus
    $self->write($select . "4E$pageno$pagedata\r");
    $y=$self->readline();           # Consume command echo
    $self->OW_Reset();
    $self->write($select . "BE00FFFFFFFFFFFFFFFFFF\r"); # Read scratchpad
    $y = $self->readline();       # Consume command echo and get data
    $data = substr($y,22,18);    # Get the scratchpad data
    $self->OW_Reset();                  # Clean things up
    $ok = $self->crc8_LR($data);        # check the CRC
    if (!$ok) {
        $self->write($select . "48$pageno\r"); # Issue copy scratchpad
        usleep(150_000);           # Wait 15ms
        $y = $self->readline();    # Consume command echo
        $y = $self->Read26_page($SerialNumber,$pageno);
        if ($y eq $pagedata) {	# Compare current contents with intended contents
            return(0);
        } else {
            return(1);
        }
    }
}

##/////////////////////////////////////////////////////////////////
#/      							  /
#/   Read28_Temp ... acquire and convert temp from DS18B20        /
#/							          /
#/   Input: SerialNumber of DS18B20 (ASCII Hex LSB = family)      /
#/                                                                /
#/   Output: Scalar containing temperature in units of TScale     /
#/                                                                /
#//////////////////////////////////////////////////////////////////
sub Read28_Temp {
    my $self = shift;
    my $SerialNum = shift;		#Get the passed-in serial number
    my ($ok, $x, $hextemp, $TempC) = 0;
    my ($line) = "";
    my $select = "b55" . $self->ReverseSN($SerialNum);

    $ok = $self->OW_Reset();

    $self->write($select . "BEFFFFFFFFFFFFFFFFFFx");
    $x = $self->readline();                        #Read echo and response
    $line = substr($x,20,20);
    $hextemp = hex((substr($line,2,2)) . substr($line,0,2));
    #//////////////////////////////
    #/   Handle negative number   /
    #//////////////////////////////
    if ($hextemp & 0x8000) {		# Is the sign bit set?
       $hextemp = $hextemp | 0xFFFF0000; # propagate the sign bit
       $hextemp = -((~$hextemp)+1);     # and negate
    }

    $TempC = $hextemp * 0.0625;
    return $self->CtoTscale($TempC);
}

sub Read10_Temp {
    my $self = shift;
    my $SerialNum = shift;    #Get the passed-in serial number
    my ($ok, $x, $hextemp, $TempC) = 0;
    my ($line) = "";
    my $select = "b55" . $self->ReverseSN($SerialNum);

    $ok = $self->OW_Reset();

    $self->write($select . "BEFFFFFFFFFFFFFFFFFFx");
    $x = $self->readline();                        #Read echo and response
    $line = substr($x,20,20);

    $hextemp = hex(substr($line,0,2));


    #//////////////////////////////
    #/   Handle negative number   /
    #//////////////////////////////
    if ($hextemp & 0x8000) {    # Is the sign bit set?
       $hextemp = $hextemp | 0xFFFF0000; # propagate the sign bit
       $hextemp = -((~$hextemp)+1);     # and negate
    }

    $TempC = $hextemp * 0.5;
    return $self->CtoTscale($TempC);
}

#////////////////////////////////////////////////////////////
#/                                                          /
#/   Expand TString ... read the DS2431 on a T-String and   /
#/      return a comma delimited list of DS18B20 serial     /
#/      numbers on the T-String.  The list is in order from /
#/      the DS2431 end of the string (RJ45 end).            /
#/                                                          /
#////////////////////////////////////////////////////////////
sub Expand_TString {
    my $self = shift;
    my $StringSN = shift;			# Get serial number passed in.
    my ($x, $count) = 0;
    my ($line, $result) = "";
    my $select = "b55" . $self->ReverseSN($StringSN);
    
    #///////////////////////////////////////////////
    #/   Get the header from the T-String EEPROM   /
    #///////////////////////////////////////////////
    $self->OW_Reset() or die;
    $x = $self->write($select . "F00000");
    $line = $self->read(24);                         # read echo from command
    #///////////////////////////////////////////////////////
    #/   Read the header from the 1st page of the DS2431   /
    #///////////////////////////////////////////////////////
    $x = $self->write("FFFFFFFFFFFFFFFF");
    $line = $self->read(16);                      # read 16 characters (8 bytes)
    #/////////////////////////////////////////////////////////////
    #/   Get up to 15 serial numbers of 18B20s from the DS2431   /
    #/   Store them in $result for later use in reading the      /
    #/   sensors in physical order from the RJ45 end.            /
    #/////////////////////////////////////////////////////////////
    $result =  "";				# clear the result string
    for ($count=0; $count<16; $count++) {
      $x = $self->write("FFFFFFFFFFFFFFFF");
      $line = $self->read(16);                      # read 16 characters (8 bytes)
      if ($line eq "0000000000000000") {
        last;
      }
      if ($count == 0) {
          $result = $line;
      } else {
          $result = $result . "," . $line;
      }
   }
   $x = $self->write("\r");                         # A carriage return to get out of byte mode
   $line = $self->read(2);                          # consume the resonse CrLf.
   return $result;
}

#////////////////////////////////////////////////////////////////////////
#/                                                                      /
#/   Read_Sensor ... A layer to shield the user from a sensor type from /
#/                   how a sensor is implemented.                       /
#/                                                                      /
#/                   A sensor with a type of "TH" returns a temperature /
#/                   in degrees C and a Humidity as a percentage        /
#/                                                                      /
#/                   Read sensor takes care of figuring out whether     /
#/                   the sensor is based upon a HIH-3610 humidity       /
#/                   sensor or a HIH-4000 (they have different          /
#/                   conversion curves).  The result is that the        /
#/                   sensor reading is returned as percentage RH in     /
#/                   either case.                                       /
#/                                                                      /
#////////////////////////////////////////////////////////////////////////                                                                 
sub Read_Sensor {
    my $self = shift;
    my $SerialNumber = shift;               # The first paramenter is 1-wire
                                           # ID (SerialNumber).
    my $x = shift;                          # Get the characteristics list
    my $y = 0;                              # Scratchpad
    my ($z, $result) = "";
    my @value = split(',',$x);              # Convert the characteristic list to an array
    my @answer=();                          # Scratchpad

   #//////////////////////////
   #/   Serial Number only   /
   #//////////////////////////
   if ($value[0] eq "SN") { return "SN"; }

   #///////////////////////////////
   #/   Temperature only sensor   /
   #///////////////////////////////
   if ($value[0] eq "T") {
      if (substr($SerialNumber,14,2) eq "28") {
         $y = $self->Read28_Temp($SerialNumber);
         $result = sprintf ("T,%.3f",$y);
         return $result;
      }
      if (substr($SerialNumber,14,2) eq "26") {
         @answer = $self->Read26_TADC($SerialNumber);
         $result = sprintf ("T,%.3f",$answer[0]);
         return $result;
      }
      if (substr($SerialNumber,14,2) eq "10") {
         @answer = $self->Read10_Temp($SerialNumber);
         $result = sprintf ("T,%.3f",$answer[0]);
         return $result;
      }      
   }
   #///////////////////////////////////////
   #/   Temperature and Humidity sensor   /
   #///////////////////////////////////////
   if ($value[0] eq "TH") {
      @answer = $self->Read26_TADC($SerialNumber);
      if(@answer)
      {
        if ($value[1] == 3610) {
           $y = ($answer[1]-0.9581)/0.030680;
        }
        if ($value[1] == 4000) {
           $y = ($answer[1]-0.826)/0.031483;
        }
        $result = sprintf ("TH,%.3f,%.1f",$answer[0],$y);
        return $result;
      }
      else
      {
        return "";
      }
   }
   #/////////////////////////////////////////////////////////////
   #/   T-String ... return from 1 to 15 temperature readings   /
   #/////////////////////////////////////////////////////////////
   if ($value[0] eq "TS") {
      @answer = @value;                  # Make a copy of characteristics array
      $result = "TS";			 # Initialize result string
      shift(@answer);			 # Drop the first element of array
      foreach $x (@answer) {
         $y = $self->Read28_Temp($x);
         $z = sprintf (",%.3f",$y);
         $result .= $z;			 # Concantonate new reading onto result 
      }
      return $result;
   }
   #////////////////////////////////////
   #/   Temperature and Light sensor   /
   #////////////////////////////////////
   if ($value[0] eq "TL") {
      @answer = $self->Read26_TADC($SerialNumber);
      $result = sprintf ("TL,%.3f,%.3f",$answer[0],$answer[1]);
      return $result;
   }
   #//////////////////////////////////////
   #/   Temperature and Voltage sensor   /
   #//////////////////////////////////////
   if ($value[0] eq "TV") {
      @answer = $self->Read26_TADC($SerialNumber);
      $result = sprintf ("TV,%.3f,%.3f",$answer[0],$answer[1]);
      return $result;
   }
   #//////////////////////////////////////
   #/   Temperature and Current sensor   /
   #//////////////////////////////////////
   if ($value[0] eq "TL") {
      @answer = $self->Read26_TADC($SerialNumber);
      $result = sprintf ("TC,%.3f,%.3f",$answer[0],$answer[1]);
      return $result;
   }

   #/////////////////////////////////////////////
   #/   Temperature and Differential Pressure   /
   #/////////////////////////////////////////////
   if ($value[0] eq "TPD") {
      @answer = $self->Read26_TADC($SerialNumber);
      $result = sprintf ("TPD,%.3f,%.3f",$answer[0],$answer[1]);
      return $result;
   }
   #///////////////////////////////////////
   #/   Temperature and Static Pressure   /
   #///////////////////////////////////////
   if ($value[0] eq "TPS") {
      @answer = $self->Read26_TADC($SerialNumber);
      $result = sprintf ("TPS,%.3f,%.3f",$answer[0],$answer[1]);
      return $result;
   }

   #/////////////////////////
   #/   Sensor is unknown   /
   #/////////////////////////
   return "UN";
}
#/////////////////////////////////////////////////////////////////////
#/                                                                   /
#/   Populate $theTime with a string of the local time in the form   /
#/   mm/dd/yyyy hh:mm:ss					     /
#/								     /
#/////////////////////////////////////////////////////////////////////
sub Get_the_time {
    my $self = shift;
    my ($second, $minute, $hour, $dayOfMonth, $month, $yearOffset, $dayOfWeek, $dayOfYear, $daylightSavings) = localtime();
    my $year = 1900 + $yearOffset;
    my $theTime = sprintf("%02u/%02u/%4u %02u:%02u:%02u",$month,$dayOfMonth,$year,$hour,$minute,$second);
    return $theTime;
}
#/////////////////////////////////////////////////////////////
#/                                                           /
#/   ConvertT                                                /
#/                                                           /
#/   Issue a Convert-T for all sensors on the bus.           /
#/                                                           /
#/////////////////////////////////////////////////////////////
sub ConvertT {
    my $self=shift;
    my $line = "";
    $self->OW_Reset() or die;         # Start off with a reset
    $self->write("bCC\r");          # Select all the sensors
    $line=$self->readline();        # Consume echo
    $self->write("p44");            # Start the convert-T with strong pullup
    usleep(900_000);                # Wait 900ms (Maximum convert time)
    $self->write("\r");             # end binary mode
    $line=$self->readline();        # Consume the command echo
    $self->OW_Reset() or die;         # End the strong pull-up
    return;
}
#/////////////////////////////////////////////////////////////
#/                                                           /
#/   ConvertV                                                /
#/                                                           /
#/   Issue a Convert-V for all sensors on the bus.           /
#/                                                           /
#/////////////////////////////////////////////////////////////
sub ConvertV {
    my $self =shift;
    my $line = "";
    $self->OW_Reset() or die;     # Start off with a reset
    $self->write("bCC\r");      # Select all the sensors
    $line=$self->readline();    # Consume echo
    $self->write("pB4");        # Start the convert-T with strong pullup
    usleep(10_000);             # Wait 900ms (Maximum convert time)
    $self->write("\r");         # end binary mode
    $line=$self->readline();    # Consume the command echo
    $self->OW_Reset() or die;     # End the strong pull-up
    return;
}

#/////////////////////////////////////////////////////////////
#/                                                           /
#/   Read26_TADC ... Read the temp and A/D from DS2438       /
#/                                                           /
#/   This routine assumes that a Convert-T and Convert-V     /
#/   has been issued prior to calling it.                    /
#/                                                           /
#/   Input: SerialNumber of DS2438 (ASCII Hex LSB = family)  /
#/   It returns a 3 element array.                           /
#/							     /
#/   Element 0 is temperature in units of Tscale             /
#/   Element 1 is A/D voltage (0-10V)			     /
#/   Element 2 is current sense voltage (0-0.25V)            /
#/                                                           /
#/////////////////////////////////////////////////////////////
sub Read26_TADC {
    my $self = shift;
    my @result = ();                    #Clear the result array
    my ($ok, $SCR) = 0;
    my ($line, $data) = '';
    my $select = "b55" . $self->ReverseSN(shift);

    # Set VAD Flag
    $ok = $self->OW_Reset();
    $self->write($select . "4E0000\r");
    $line = $self->readline();

    $ok = $self->OW_Reset();
    $self->write($select . "4800\r");
    $line = $self->readline();

    $ok = $self->OW_Reset();
    $self->write($select . "B800\r");	#Recall to scratchpad
    $line = $self->readline();	# Consume command echo

    $ok = $self->OW_Reset();
    $self->write($select . "BE00FFFFFFFFFFFFFFFFFF\r");
    $line = $self->readline();         #retrieve response w/2 second timeout

    $data = substr($line,22,18);
    $ok = $self->crc8_LR(substr($data,0,18));   # Validate CRC on data read
    if ($ok != 0) {
       print "CRC on data read from DS2438 substr($select,3,16)\n";
       return undef;
    }
    $SCR = hex(substr($data,0,2));    #Change the SCR to binary
    if (($SCR & 0x10)) {
       print "Temperature conversion was busy when reading taken\n";
       return undef;
    }
    if (($SCR & 0x40)) {
       print "A/D conversion was busy when reading taken\n";
       return undef;
    }
    if (($SCR & 0x08)) {
       print "A/D wrong input ... voltage is Vdd\n";
       return undef;
    }

    my $hextemp = hex(substr($data,4,2) . substr($data,2,2)); #Swap MSB/LSB
    #//////////////////////////////
    #/   Handle negative number   /
    #//////////////////////////////
    if ($hextemp & 0x8000) {		# Is the sign bit set?
       $hextemp = $hextemp | 0xFFFF0000; # propagate the sign bit
       $hextemp = -((~$hextemp)+1);     # and negate
    }

    $hextemp = $hextemp >> 3;			#Get the stars to line up

    $result[0] = $hextemp * 0.03125;
    $result[0] = $self->CtoTscale($result[0]);
    
    $result[1] = (hex(substr($data,8,2).substr($data,6,2))) * 0.01;

    $hextemp = hex(substr($data,12,2) . substr($data,10,2));
    #//////////////////////////////
    #/   Handle negative voltage   /
    #//////////////////////////////
    if ($hextemp & 0x8000) {		# Is the sign bit set?
       $hextemp = $hextemp | 0xFFFF0000; # propagate the sign bit
       $hextemp = -((~$hextemp)+1);     # and negate
    }

    $result[2] = $hextemp * 0.0002441;
    return @result;
}

#/////////////////////////////////////////////////////////////
#/                                                           /
#/   Sensor_ConvertT                                         /
#/                                                           /
#/   Issue a Convert-T for a specific sensor and wait for    /
#/   completion.                                             /
#/                                                           /
#/   Input: SerialNumber of sensor (ASCII Hex LSB = family)  /
#/                                                           /
#/////////////////////////////////////////////////////////////
sub Sensor_ConvertT {
    my $self=shift;
    my $line="";
    my $select = self->ReverseSN(shift);
    $self->OW_Reset() or die;        # Start off with a reset
    $self->write("$select\r");       #Select the sensor
    $self->write("p44");             # Start the convert-T with strong pullup
    $self->write("\r");              # end binary mode and strong pullup
    if(substr($select,17,2) eq "28") { usleep(900000); }
    if(substr($select,17,2) eq "10") { usleep(900000); }
    if(substr($select,17,2) eq "26") { usleep(10000); }
    $line=IBLreadline();             # Consume the command echo
    $self->OW_Reset() or die;        # End the strong pull-up and binary mode
    return;
}
#/////////////////////////////////////////////////////////////
#/                                                           /
#/   Sensor_ConvertV                                         /
#/                                                           /
#/   Issue a Convert-V for a specific sensor and wait for    /
#/   completion.                                             /
#/                                                           /
#/   Input: SerialNumber of sensor (ASCII Hex LSB = family)  /
#/                                                           /
#/////////////////////////////////////////////////////////////
sub Sensor_ConvertV {
    my $self=shift;
    my $line;
    my $select = self->ReverseSN(shift);
    $self->OW_Reset() or die;        # Start off with a reset
    $self->write("$select\r");       # Select the sensor
    $self->write("pB4");             # Start the convert-V with strong pullup
    $self->write("\r");              # end binary mode
    if(substr($select,17,2) eq "26") { usleep(10000); }
    $line=$self->readline();         # Consume the command echo
    $self->OW_Reset() or die;        # End the strong pull-up
    return;
}
##///////////////////////////////////////////////////////////////////////
##/                                                                     /
##/   crc8_LR ... calculate and/or compare CRC8 from Left to Right of   /
##/               string passed in                                      /
##/  									/
##/   The format of the string passed in is:				/
##/   1. A hex character string with the transmitted CRC in the         /
##/      right hand position (data from a sensor)                       /
##/   2. Each pair of hex ASCII characters represents 1 byte of the     /
##/      data.                                                          /
##/									/
##/   $crc8 (the returned value) should be 0 if the input string        /
##/    had a CRC in the leftmost position, else, $crc8 is the crc of    /
##/   the string.                                                       /
##/									/
##///////////////////////////////////////////////////////////////////////
sub crc8_LR  {
    my $self=shift;
    my $string = shift;
    my ($ycounter,$ylen,$yloop,$yy,$crc8)=0;
    $crc8 = 0; $ycounter = 0;
    $ylen = length($string); $yloop = $ylen/2;
    while($ycounter < $yloop) {
        $yy = hex(substr($string,($ycounter*2),2));    #Read the next rightmost 2 characters
        $crc8 = $crc8_table[$yy ^ $crc8];
        $ycounter++;                              #increment $ycounter by 1
    }
  return $crc8;
}
##///////////////////////////////////////////////////////////////////////
##/                                                                     /
##/   crc8_RL ... calculate and/or compare CRC8 from Right to Left of 	/
##/               string passed in                                      /
##/  									/
##/   The format of the string passed in is:				/
##/   1. A hex character string with the transmitted CRC in the         /
##/      left hand position (family code on right.                      /
##/   2. Each pair of hex ASCII characters represents 1 byte of the     /
##/      data.                                                          /
##/									/
##/   $crc8 (the returned value) should be 0 if the input string        /
##/    had a CRC in the leftmost position, else, $crc8 is the crc of    /
##/   the string.                                                       /
##/									/
##///////////////////////////////////////////////////////////////////////
sub crc8_RL  {
    my $self = shift;
    my $string = shift;
    my ($ycounter, $ylen, $yloop, $yy, $crc8)=0;
    $crc8 = 0; $ycounter = 0;
    my $zz = 0;
    $ylen = length($string); $yloop = $ylen/2;
    while($ycounter < $yloop) {
        $yy = hex(substr($string,($ylen-($ycounter*2)-2),2));	#Read the next rightmost 2 characters
        $zz = $yy ^ $crc8;
        $crc8 = $crc8_table[$yy ^ $crc8];
        $ycounter++;                                        #decrement $ycounter by 2
    }
return $crc8;
}

#///////////////////////////////////
#/				   /
#/   DSIH ... Dump string in Hex   /
#/				   /
#///////////////////////////////////
sub DSIH {
    my $self = shift;
    my $x = shift;
    my $z = "";
    print STDERR "String dump: ";
    for (0..length($x)) {
        $z = ord(substr($x,$_,1));
        printf STDERR "%2.2X ",$z;
    }
print STDERR "\n";
}
#//////////////////////////////////////////////////////////////////////////////
#/                                                                            /
#/   Convert temperature from degrees C to units associates with TEMP_SCALE   /
#/                                                                            /
#//////////////////////////////////////////////////////////////////////////////
sub CtoTscale {
    my $self = shift;
    my $temp = shift;
    if ($self->{TEMP_SCALE} eq "C") { return $temp}
    if ($self->{TEMP_SCALE} eq "F") { return (($temp*9)/5.0)+32.0 }
    if ($self->{TEMP_SCALE} eq "K") { return ($temp + 273.0) }
}

1;
